/*----------------------------------------------------------------------
  File    : xhfn.c
  Contents: Hopfield network as associative memory
  Author  : Christian Borgelt
  History : 2002.12.01 file created from file xsom.c
            2003.01.22 mouse event handling improved
            2013.08.09 adapted to higher compiler warning level
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Core.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>
#include "menu.h"
#include "dialog.h"
#include "fselect.h"
#include "random.h"
#include "hopfnet.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- error codes --- */
#define E_NONE         0        /* no error */
#define E_NOMEM      (-1)       /* not enough memory */
#define E_FOPEN      (-2)       /* cannot open file */
#define E_FREAD      (-3)       /* read error on file */
#define E_FWRITE     (-4)       /* write error on file */
#define E_INIT       (-5)       /* initialization failed */
#define E_WIDGET     (-6)       /* widget creation failed */
#define E_DIALOG     (-7)       /* cannot create dialog box */
#define E_COLOR      (-8)       /* color allocation failed */
#define E_GRAPH      (-9)       /* cannot create graphics context */
#define E_PATCNT    (-10)       /* pattern capacity exceeded */
#define E_UNKNOWN   (-11)       /* unkown error */

#define FMT_MAX      256        /* max. len. of error message format */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- error message data --- */
  const char *name, *type;      /* name and type of the error message */
  const char *class;            /* resource class of error message */
  const char *dflt;             /* default message */
  int        argc;              /* number of arguments */
} ERRMSG;                       /* (error message) */

/*----------------------------------------------------------------------
  Action Function Prototypes
----------------------------------------------------------------------*/
static void mouse  (Widget w, XEvent *e, String *s, Cardinal *c);
static void load   (Widget w, XEvent *e, String *s, Cardinal *c);
static void save   (Widget w, XEvent *e, String *s, Cardinal *c);
static void quit   (Widget w, XEvent *e, String *s, Cardinal *c);
static void clear  (Widget w, XEvent *e, String *s, Cardinal *c);
static void store  (Widget w, XEvent *e, String *s, Cardinal *c);
static void delete (Widget w, XEvent *e, String *s, Cardinal *c);
static void next   (Widget w, XEvent *e, String *s, Cardinal *c);
static void init   (Widget w, XEvent *e, String *s, Cardinal *c);
static void exec   (Widget w, XEvent *e, String *s, Cardinal *c);
static void redraw (Widget w, XEvent *e, String *s, Cardinal *c);
static void grid   (Widget w, XEvent *e, String *s, Cardinal *c);
static void params (Widget w, XEvent *e, String *s, Cardinal *c);
static void about  (Widget w, XEvent *e, String *s, Cardinal *c);

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
#include "xhfn.rsc"             /* fallback resources */

static XtActionsRec actions[] = {
  { "mouse",    mouse    },     /* mouse movement */
  { "load",     load     },     /* File     > Load Map... */
  { "save",     save     },     /* File     > Save Map... */
  { "quit",     quit     },     /* File     > Quit */
  { "clear",    clear    },     /* Actions  > Clear */
  { "next",     next     },     /* Actions  > Next Pattern */
  { "store",    store    },     /* Actions  > Store Pattern */
  { "delete",   delete   },     /* Actions  > Delete Pattern */
  { "init",     init     },     /* Actions  > Initialize */
  { "redraw",   redraw   },     /* Actions  > Redraw */
  { "exec",     exec     },     /* Actions  > Start/Stop Training */
  { "grid",     grid     },     /* Settings > Grid... */
  { "params",   params   },     /* Settings > Parameters... */
  { "about",    about    },     /* Help     > About XSOM... */
  { "db_focus", db_focus },     /* set focus to input gadget */
  { "db_next",  db_next  },     /* switch to next input gadget */
  { "db_close", db_close } };   /* close dialog box */

/* --- error messages --- */
static const ERRMSG errmsgs[] = {   /* error message data */
  /* E_NONE      0 */  { "error",   "none",    "XHfN.Error",
                         "no error",                               0 },
  /* E_NOMEM    -1 */  { "error",   "nomem",   "XHfN.Error",
                         "not enough memory",                      0 },
  /* E_FOPEN    -2 */  { "warning", "fopen",   "XHfN.Warning",
                         "cannot open file:\n%s",                  1 },
  /* E_FREAD    -3 */  { "warning", "fread",   "XHfN.Warning",
                         "read error on file:\n%s",                1 },
  /* E_FWRITE   -4 */  { "warning", "fwrite",  "XHfN.Warning",
                         "write error on file:\n%s",               1 },
  /* E_INIT     -5 */  { "error",   "init",    "XHfN.Error",
                         "initialization failed",                  0 },
  /* E_WIDGET   -6 */  { "error",   "widget",  "XHfN.Error",
                         "widget creation failed",                 0 },
  /* E_DIALOG   -7 */  { "warning", "dialog",  "XHfN.Warning",
                         "cannot create dialog box",               0 },
  /* E_COLOR    -8 */  { "warning", "color",   "XHfN.Warning",
                         "color allocation failed",                0 },
  /* E_GRAPH    -9 */  { "error",   "gc",      "XHfN.Error",
                         "cannot create graphics context",         0 },
  /* E_PATCNT  -10 */  { "warning", "patcnt",  "XHfN.Warning",
                         "pattern capacity exceeded",              0 },
  /* E_UNKNOWN -11 */  { "error", "unknown",   "XHfN.Error",
                         "unknown error",                          0 }};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static XtAppContext appctx;     /* X11 application context */
static Display *display;        /* display connected to */
static int     screen;          /* screen of the display */
static Window  window;          /* main window */
static Atom    wm_delwin;       /* delete window atom of window mgr. */
static char    fmt[FMT_MAX];    /* format and buffer for messages */
static char    buf[PATH_MAX +FMT_MAX];

/* --- main widgets and menu --- */
static Widget  w_top;           /* top level widget */
static Widget  w_main;          /* main window widget */
static Widget  w_view;          /* viewport widget */
static Widget  w_net;           /* network widget */
static MENU    *menu;           /* pull down menu */
static GC      gc;              /* graphics context for drawing */

/* --- property sheets --- */
static PSHEET  *psh_grid   = NULL;  /* grid and shape */
static PSHEET  *psh_params = NULL;  /* learning parameters */

/* --- color information --- */
static unsigned long black;     /* black and white */
static unsigned long white;     /* for normal drawing */
static unsigned long grey;      /* grey for background shape */

/* --- neural network information --- */
static struct {                 /* -- parameters -- */
  int    width, height;         /* width and height of network */
  int    grid;                  /* grid width (in pixels) */
  int    mode;                  /* update mode */
  int    curr;                  /* current pattern */
  double delay;                 /* delay (in seconds) */
} nnpar = { 10, 12, 24, 0, -1, 0.01 };

static HOPFNET *hfn    = NULL;  /* Hopfield network */
static int     running = 0;     /* whether computations are running */
static XtIntervalId iid;        /* id of time interval procedure */

/*----------------------------------------------------------------------
  Error Reporting Functions
----------------------------------------------------------------------*/

static void dcb_nop (Widget widget, XtPointer client, XtPointer call)
{                               /* --- close dialog box (callback) */
  while (!XtIsTransientShell(widget))
    widget = XtParent(widget);  /* find the shell widget */
  XtPopdown(widget);            /* close the dialog box */
  mn_enable(menu, MN_MENU, 1);  /* and enable the menu bar */
}  /* dcb_nop() */

/*--------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list      va;              /* list of variable arguments */
  const ERRMSG *msg;            /* error message */
  String       argv[8];         /* variable argument vector */
  Cardinal     argc;            /* number of arguments */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs -code;          /* get error message data */
  if (!msg->name) msg = errmsgs -E_UNKNOWN;
  va_start(va, code);           /* get list of variable arguments */
  for (argc = 0; (int)argc < msg->argc; argc++)
    argv[argc] = va_arg(va, String);
  va_end(va);                   /* get message arguments */
  if (strcmp(msg->name, "error") == 0)  /* if name is 'error' */
    XtAppErrorMsg(appctx, msg->name, msg->type, msg->class,
      msg->dflt, argv, &argc);  /* print error message */
  else {                        /* if name is 'warning' */
    XtAppGetErrorDatabaseText(appctx, msg->name, msg->type,
      msg->class, msg->dflt, fmt, FMT_MAX, NULL);
    va_start(va, code);         /* list of variable arguments */
    vsprintf(buf, fmt, va);     /* get and format message */
    va_end(va);                 /* end variable argument evaluation */
    mn_enable(menu, MN_MENU,0); /* disable menu bar and show dialog */
    if (db_alert(w_top, dcb_nop, buf) != 0) {
      mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
      XtAppWarningMsg(appctx, msg->name, msg->type, msg->class,
        msg->dflt, argv, &argc);/* if the alert dialog box */
    }                           /* cannot be shown, */
  }                             /* print a warning message */
}  /* error() */

/*--------------------------------------------------------------------*/

static void resize (int width, int height)
{                               /* --- resize widgets */
  Dimension w, h;               /* width and height of top widget */

  width  *= nnpar.grid;         /* compute the extensions */
  height *= nnpar.grid;         /* in pixels */
  w = mn_minwidth(menu)-1;      /* compute board size in pixels and */
  h = (Dimension)(mn_height(menu) +height);   /* height of interior */
  if (width > w) w = (Dimension)width;        /* get minimal width */
  XtVaSetValues(w_top,  XtNwidth,    w+1, XtNheight,    h+1,
                        XtNmaxWidth, w+1, XtNmaxHeight, h+1,      NULL);
  XtVaSetValues(w_net,  XtNwidth,    w-1, XtNheight,    height-1, NULL);
  XtVaSetValues(w_view, XtNwidth,    w-1, XtNheight,    height-1, NULL);
  mn_resize(menu, w);           /* set new widget extensions */
}  /* resize() */

/*----------------------------------------------------------------------
  Dialog Box Callback Functions
----------------------------------------------------------------------*/

static void dcb_load (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Load Map...' callback */
  FILE    *file;                /* output file */
  HOPFNET *tmp;                 /* temp. buffer for Hopfield network */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  file = fopen(fname, "r");     /* open the input file */
  if (!file) { error(E_FOPEN, fname); return; }
  tmp = hfn_load(file);         /* load the Hopfield network */
  fclose(file);                 /* close the input file */
  if (!tmp) { error(E_FREAD, fname); return; }
  hfn_delete(hfn); hfn = tmp;   /* replace the existing network */
  nnpar.width  = hfn_width(hfn);
  nnpar.height = hfn_height(hfn);
  resize(nnpar.width, nnpar.height);
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_load() */

/*--------------------------------------------------------------------*/

static void dcb_save (Widget widget, XtPointer ok, XtPointer fname)
{                               /* --- 'Save Map...' callback */
  FILE *file;                   /* output file */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  file = fopen(fname, "w");     /* open the output file */
  if (!file) { error(E_FOPEN, fname); return; }
  hfn_save(hfn, file);          /* save the Hopfield network */
  if (fclose(file) != 0)        /* (list of stored patterns), */
    error(E_FWRITE);            /* then close the output file */
}  /* dcb_save() */

/*--------------------------------------------------------------------*/

static void dcb_grid (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Grid...' callback */
  HOPFNET *tmp;                 /* temp. buffer for Hopfield network */
  int     w;                    /* buffer for widths */

  mn_enable(menu, MN_MENU, 1);  /* enable the menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  nnpar.width  = atoi(psh_getval(psh_grid, "width"));
  if      (nnpar.width  < 2)   nnpar.width  = 2;
  else if (nnpar.width  > 100) nnpar.width  = 100;
                                /* width  of network (in units) */
  nnpar.height = atoi(psh_getval(psh_grid, "height"));
  if      (nnpar.height < 1)   nnpar.height = 1;
  else if (nnpar.height > 100) nnpar.height = 100;
                                /* height of network (in units) */
  nnpar.grid   = atoi(psh_getval(psh_grid, "grid"));
  if      (nnpar.grid   < 4)   nnpar.grid   = 4;
  else if (nnpar.grid   > 64)  nnpar.grid   = 64;
                                /* grid width (in pixels) */
  tmp = hfn_create(nnpar.width, nnpar.height);
  if (!tmp) { error(E_NOMEM); return; }
  hfn_delete(hfn); hfn = tmp;   /* replace the Hopfield network */
  w = mn_minwidth(menu);        /* and check the grid width */
  if (w > nnpar.width *nnpar.grid)
    nnpar.grid = (w +nnpar.width -1) /nnpar.width;
  resize(nnpar.width, nnpar.height); /* resize the window and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_grid() */

/*--------------------------------------------------------------------*/

static void dcb_params (Widget widget, XtPointer ok, XtPointer call)
{                               /* --- 'Parameters...' callback */
  mn_enable(menu, MN_MENU, 1);  /* enable menu bar */
  if (!ok) return;              /* if not ok pressed, abort */
  nnpar.mode  = (int)(ptrdiff_t)psh_getval(psh_params, "mode");
                                /* neuron update mode */
  nnpar.delay = atof(psh_getval(psh_params, "delay"));
  if      (nnpar.delay < 0.001) nnpar.delay  = 0.001;
  else if (nnpar.delay > 10)    nnpar.delay  = 10;
                                /* delay between redraws */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* dcb_params() */

/*----------------------------------------------------------------------
  Menu Callback Functions
----------------------------------------------------------------------*/

static void mcb_load   (Widget widget, XtPointer client, XtPointer call)
{ load(NULL, NULL, NULL, NULL); }

static void mcb_save   (Widget widget, XtPointer client, XtPointer call)
{ save(NULL, NULL, NULL, NULL); }

static void mcb_quit   (Widget widget, XtPointer client, XtPointer call)
{ quit(NULL, NULL, NULL, NULL); }

static void mcb_clear  (Widget widget, XtPointer client, XtPointer call)
{ clear(NULL, NULL, NULL, NULL); }

static void mcb_next   (Widget widget, XtPointer client, XtPointer call)
{ next(NULL, NULL, NULL, NULL); }

static void mcb_store  (Widget widget, XtPointer client, XtPointer call)
{ store(NULL, NULL, NULL, NULL); }

static void mcb_delete (Widget widget, XtPointer client, XtPointer call)
{ delete(NULL, NULL, NULL, NULL); }

static void mcb_init   (Widget widget, XtPointer client, XtPointer call)
{ init(NULL, NULL, NULL, NULL); }

static void mcb_start  (Widget widget, XtPointer client, XtPointer call)
{ exec(NULL, NULL, NULL, NULL); }

static void mcb_redraw (Widget widget, XtPointer client, XtPointer call)
{ redraw(NULL, NULL, NULL, NULL); }

static void mcb_grid   (Widget widget, XtPointer client, XtPointer call)
{ grid(NULL, NULL, NULL, NULL); }

static void mcb_params (Widget widget, XtPointer client, XtPointer call)
{ params(NULL, NULL, NULL, NULL); }

static void mcb_about  (Widget widget, XtPointer client, XtPointer call)
{ about(NULL, NULL, NULL, NULL); }

/*----------------------------------------------------------------------
  Action Functions
----------------------------------------------------------------------*/

static void mouse (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- evaluate mouse button */
  int x, y;                     /* mouse/neuron coordinates */
  int g;                        /* grid width */
  static int a = 1;             /* activation to set */

  if (running) return;          /* if update is running, abort */
  if (*c <= 0) return;          /* if no argument is given, abort */
  x = e->xbutton.x;             /* get the position at which */
  y = e->xbutton.y;             /* the mouse button was pressed */
  g = nnpar.grid;               /* compute the neuron coordinates */
  x /= g; if ((x < 0) || (x >= hfn_width(hfn)))  return;
  y /= g; if ((y < 0) || (y >= hfn_height(hfn))) return;
  if (s[0][0] == '1') {         /* on mouse down, get activation */
    a = (hfn_actget(hfn, x, y) > 0) ? -1 : 1; }
  hfn_actset(hfn, x, y, a);     /* toggle the activation and redraw */
  XSetForeground(display, gc, (a > 0) ? grey : white);
  XFillRectangle(display, window, gc, x*g, y*g,
                (unsigned)(g-1), (unsigned)(g-1));
  XFlush(display);              /* flush buffered drawing commands */
}  /* mouse() */

/*--------------------------------------------------------------------*/

static void load (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Load Map... */
  const char *t;                /* buffer for dialog box title */

  if (running) return;          /* if update is running, abort */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*load"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_load) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* load() */                 /* open a file select dialog */

/*--------------------------------------------------------------------*/

static void save (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Save Map... */
  const char *t;                /* buffer for dialog box title */

  if (running) return;          /* if update is running, abort */
  mn_enable(menu, MN_MENU, 0);  /* disable the menu bar */
  XtVaGetValues(XtNameToWidget(mn_menubar(menu), "*save"),
    XtNlabel, &t, NULL);        /* get the name of the menu entry */
  if (fselect(w_top, t, NULL, NULL, dcb_save) != 0) {
    error(E_DIALOG); mn_enable(menu, MN_MENU, 1); }
}  /* save() */                 /* open a file select dialog */

/*--------------------------------------------------------------------*/

static void quit (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- File > Quit */
  if (hfn) hfn_delete(hfn);     /* delete a Hopfield network */
  if (w_top) XtDestroyApplicationContext(appctx);
  exit(0);                      /* terminate the program */
}  /* quit() */

/*--------------------------------------------------------------------*/

static void clear (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Clear */
  hfn_clear(hfn);               /* clear all activations and */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* clear() */

/*--------------------------------------------------------------------*/

static void next (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Next Pattern */
  int n;                        /* number of patterns */

  if (running) return;          /* if update is running, abort */
  n = hfn_patcnt(hfn);          /* get and check */
  if (n <= 0) return;           /* the number of patterns */
  if (++nnpar.curr >= n) nnpar.curr = 0;
  hfn_patget(hfn, nnpar.curr);  /* get the current pattern */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* next() */

/*--------------------------------------------------------------------*/

static void store (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Store Pattern */
  if (running) return;          /* if update is running, abort */
  if (hfn_patadd(hfn) != 0)     /* store the current pattern */
    error(E_PATCNT);            /* and check for an error */
}  /* store() */

/*--------------------------------------------------------------------*/

static void delete (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Delete Pattern */
  int i;                        /* pattern index */

  if (running) return;          /* if update is running, abort */
  i = hfn_patfind(hfn);         /* find and delete the pattern */
  if (i >= 0) hfn_patrem(hfn, i);
}  /* delete() */

/*--------------------------------------------------------------------*/

static void init (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Initialize */
  hfn_rand(hfn, drand);         /* init. the activations randomly */
  redraw(NULL,NULL,NULL,NULL);  /* redraw the window contents */
}  /* init() */

/*--------------------------------------------------------------------*/

static void update (XtPointer data, XtIntervalId *id)
{                               /* --- do one training step */
  int x, y, a;                  /* neuron coordinates, activation */
  int g;                        /* grid width for display */

  if (!running) return;         /* check the update flag */
  if (nnpar.mode == 0) {        /* if in single neuron mode */
    hfn_curr(hfn, &x, &y);      /* get the neuron coordinates */
    a = hfn_update(hfn, -1, 0); /* and update the neuron */
    g = nnpar.grid;             /* redraw the corresp. field */
    XSetForeground(display, gc, (a > 0) ? grey : white);
    XFillRectangle(display, window, gc, x*g, y*g,
                  (unsigned)(g-1), (unsigned)(g-1));
    XFlush(display); }          /* flush buffered drawing commands */
  else {                        /* if in epoch mode */
    hfn_update(hfn, -1, -1);    /* update all neurons */
    redraw(NULL, NULL, NULL, NULL);
  }                             /* redraw the window contents */
  iid = XtAppAddTimeOut(appctx, (unsigned long)(nnpar.delay *1000),
          update, NULL);        /* reregister the interval procedure */
}  /* update() */

/*--------------------------------------------------------------------*/

static void exec (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Start/Stop Training */
  if (running) {                /* if running, remove procedure */
    mn_visible(menu, MN_ITEMS, 3);
    mn_enable (menu, MN_ITEMS, 3);
    running = 0; XtRemoveTimeOut(iid); }
  else {                        /* otherwise start training */
    mn_visible(menu, MN_ITEMS, 5);
    mn_enable (menu, MN_ITEMS, 4);
    hfn_shuffle(hfn, drand);    /* shuffle the neurons */
    running = 1; update(NULL, NULL);
  }
}  /* exec() */

/*--------------------------------------------------------------------*/

static void redraw (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Actions > Redraw */
  static Dimension owd = 0;     /* old window width */
  static Dimension oht = 0;     /* old window height */
  XEvent     de;                /* dummy event */
  XRectangle rc;                /* clipping rectangle */
  Dimension  wd, ht;            /* window extensions */
  int        x, y;              /* loop variables */
  int        g, a;              /* grid width, activation */

  /* --- initialize graphics --- */
  if (!gc) {                    /* if there is no graphics context */
    gc = XCreateGC(display, window, 0, NULL);
    if (!gc) { error(E_GRAPH); return; }
  }                             /* create a graphics context */
  XtVaGetValues(w_net, XtNwidth, &wd, XtNheight, &ht, NULL);
  if (e && (e->type == Expose)  /* if the window was exposed, */
  &&  (wd == owd) && (ht == oht)) {        /* but not resized */
    rc.x      = (short)e->xexpose.x;       /* build and set a */
    rc.y      = (short)e->xexpose.y;       /* clipping rectangle */
    rc.width  = (unsigned short)e->xexpose.width;
    rc.height = (unsigned short)e->xexpose.height;
    XSetClipRectangles(display, gc, 0, 0, &rc, 1, Unsorted); }
  else {                        /* if the window was resized */
    rc.x = rc.y = 0;            /* or this is a forced redraw, */
    rc.width  = wd;             /* redraw all of the window */
    rc.height = ht;             /* (set full window rectangle) */
    while (XCheckTypedWindowEvent(display, window, Expose, &de));
  }                             /* remove all exposure events */

  /* --- draw Hopfield network --- */
  g = nnpar.grid;               /* get the grid width */
  for (y = hfn_height(hfn); --y >= 0; ) {
    for (x = hfn_width(hfn); --x >= 0; ) {
      a = hfn_actget(hfn, x,y); /* traverse the neurons */
      XSetForeground(display, gc, (a > 0) ? grey : white);
      XFillRectangle(display, window, gc, x*g, y*g,
                    (unsigned)(g-1), (unsigned)(g-1));
    }                           /* color each field according to */
  }                             /* the activation of the neuron */
  XSetForeground(display, gc, black);
  y = hfn_height(hfn) *g -1;    /* draw horizontal grid line */
  for (a = -1, x = hfn_width(hfn);  --x > 0; ) {
    a += g; XDrawLine(display, window, gc, a, 0, a, y); }
  x = hfn_width(hfn) *g -1;     /* draw vertical   grid line */
  for (a = -1, y = hfn_height(hfn); --y > 0; ) {
    a += g; XDrawLine(display, window, gc, 0, a, x, a); }

  XFlush(display);              /* flush buffered drawing commands */
  XSetClipMask(display, gc, None);  /* and clear the clipping mask */
}  /* redraw() */

/*--------------------------------------------------------------------*/

static void grid (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Grid/Shape... */
  char buf[32];                 /* temporary buffer */

  if (!psh_grid) {              /* if property sheet not created yet */
    psh_grid = psh_create("gridDlg", w_top, 1);
    if (!psh_grid) return;      /* create a property sheet */
    if ((psh_addline(psh_grid, "lwidth")              != 0)
    ||  (psh_additem(psh_grid, "width",  PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_grid, "lheight")             != 0)
    ||  (psh_additem(psh_grid, "height", PSH_EDIT, 0) != 0)
    ||  (psh_addline(psh_grid, "lgrid")               != 0)
    ||  (psh_additem(psh_grid, "grid",   PSH_EDIT, 0) != 0)) {
      psh_delete(psh_grid); error(E_DIALOG); return; }
  }                             /* add dialog items */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  sprintf(buf, "%d", nnpar.width);   /* network width  (in units) */
  psh_setval(psh_grid, "width",  buf);
  sprintf(buf, "%d", nnpar.height);  /* network height (in units) */
  psh_setval(psh_grid, "height", buf);
  sprintf(buf, "%d", nnpar.grid);    /* grid width (in pixels) */
  psh_setval(psh_grid, "grid", buf);
  psh_handle(psh_grid, dcb_grid);    /* handle the dialog */
}  /* grid() */

/*--------------------------------------------------------------------*/

static void params (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- Settings > Parameters... */
  char buf[32];                 /* buffer for numbers */

  if (!psh_params) {            /* if property sheet not created yet */
    psh_params = psh_create("paramDlg", w_top, 1);
    if (!psh_params) return;     /* create a property sheet */
    if ((psh_addline(psh_params, "lmode")               != 0)
    ||  (psh_additem(psh_params, "mode",   PSH_MENU, 0) != 0)
    ||  (psh_menuadd(psh_params, "mode", "single")      != 0)
    ||  (psh_menuadd(psh_params, "mode", "all")         != 0)
    ||  (psh_addline(psh_params, "ldelay")              != 0)
    ||  (psh_additem(psh_params, "delay",  PSH_EDIT, 0) != 0)) {
      psh_delete(psh_params); error(E_DIALOG); return; }
  }                             /* set browse button callback */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  psh_setval(psh_params, "mode", (char*)(ptrdiff_t)nnpar.mode);
  sprintf(buf, "%g", nnpar.delay);      /* activation update mode */
  psh_setval(psh_params, "delay", buf); /* delay between redraws */
  psh_handle(psh_params, dcb_params);   /* handle the dialog */
}  /* params() */

/*--------------------------------------------------------------------*/

static void about (Widget w, XEvent *e, String *s, Cardinal *c)
{                               /* --- 'About' dialog box */
  mn_enable(menu, MN_MENU, 0);  /* disable menu bar */
  if (db_about(w_top, 0, dcb_nop) != 0)
    error(E_DIALOG);            /* show 'about' dialog box */
}  /* about() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main program */
  Arg      args[8];             /* widget arguments */
  Cardinal n;                   /* argument counter */
  int      w;                   /* minimal width */
  Colormap colmap;              /* color map of display */
  XColor   xc[9];               /* buffer for colors */

  XtSetLanguageProc(NULL, (XtLanguageProc)NULL, NULL);

  /* --- create main window --- */
  w_top = XtAppInitialize(&appctx, "XHfN",
            NULL, 0, &argc, argv, fallback, NULL, 0);
  if (!w_top) error(E_INIT);    /* create toplevel widget */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNdefaultDistance, 0);    n++;
  XtSetArg(args[n], XtNresizable,       True); n++;
  w_main = XtCreateManagedWidget("main", formWidgetClass,
             w_top, args, n);   /* create main widget */
  if (!w_main) error(E_WIDGET); /* (form for window layout) */
  menu = mn_create(w_main);     /* create a menu bar and */
  if (!menu) error(E_NOMEM);    /* add titles and items */
  if ((mn_addtitle(menu, "file",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "load",     1, mcb_load,   0) != 0)
  ||  (mn_additem (menu, "save",     0, mcb_save,   0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "quit",     0, mcb_quit,   0) != 0)
  ||  (mn_addtitle(menu, "actions",  0, 0,          0) != 0)
  ||  (mn_additem (menu, "clear",    0, mcb_clear,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "next",     1, mcb_next,   0) != 0)
  ||  (mn_additem (menu, "store",    1, mcb_store,  0) != 0)
  ||  (mn_additem (menu, "delete",   1, mcb_delete, 0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "init",     0, mcb_init,   0) != 0)
  ||  (mn_additem (menu, "start",    2, mcb_start,  0) != 0)
  ||  (mn_additem (menu, "stop",     4, mcb_start,  0) != 0)
  ||  (mn_additem (menu, NULL,       0, 0,          0) != 0)
  ||  (mn_additem (menu, "redraw",   0, mcb_redraw, 0) != 0)
  ||  (mn_addtitle(menu, "settings", 0, 0,          0) != 0)
  ||  (mn_additem (menu, "grid",     1, mcb_grid,   0) != 0)
  ||  (mn_additem (menu, "params",   0, mcb_params, 0) != 0)
  ||  (mn_addtitle(menu, "help",     0, 0,          0) != 0)
  ||  (mn_additem (menu, "about",    0, mcb_about,  0) != 0)
  ||  (mn_addtitle(menu, NULL,       0, 0,          0) != 0))
    error(E_WIDGET);
  mn_visible(menu, MN_ITEMS, 3);/* show/hide      menu items */
  mn_enable (menu, MN_ITEMS, 3);/* enable/disable menu items */

  /* --- create board viewport --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNfromVert,  mn_menubar(menu)); n++;
  XtSetArg(args[n], XtNborderWidth, 1);           n++;
  XtSetArg(args[n], XtNleft,      XtChainLeft);   n++;
  XtSetArg(args[n], XtNtop,       XtChainTop);    n++;
  XtSetArg(args[n], XtNright,     XtChainRight);  n++;
  XtSetArg(args[n], XtNbottom,    XtChainBottom); n++;
  XtSetArg(args[n], XtNresizable, True);          n++;
  w_view = XtCreateManagedWidget("view",
             viewportWidgetClass, w_main, args, n);
  if (!w_view) error(E_WIDGET); /* create viewport */

  /* --- create network widget --- */
  n = 0;                        /* build argument list */
  XtSetArg(args[n], XtNborderWidth, 0);             n++;
  XtSetArg(args[n], XtNleft,        XtChainLeft);   n++;
  XtSetArg(args[n], XtNtop,         XtChainTop);    n++;
  XtSetArg(args[n], XtNright,       XtChainRight);  n++;
  XtSetArg(args[n], XtNbottom,      XtChainBottom); n++;
  XtSetArg(args[n], XtNresizable,   True);          n++;
  w_net = XtCreateManagedWidget("network",
             coreWidgetClass, w_view, args, n);
  if (!w_net) error(E_WIDGET);  /* create network */

  /* --- set widget extensions --- */
  w = mn_minwidth(menu);        /* set minimal width and height */
  XtVaSetValues(w_top, XtNminWidth,  (w > 8) ? w : 8,
                       XtNminHeight, mn_height(menu) +4, NULL);
  resize(nnpar.width, nnpar.height); /* set the window sizes */
  XtOverrideTranslations(w_net,      /* install redraw function */
    XtParseTranslationTable("<Expose>: redraw()"));

  /* --- install actions, accelerators and translations --- */
  XtAppAddActions(appctx, actions, XtNumber(actions));
  XtInstallAllAccelerators(w_main, w_main);

  /* --- create a Hopfield network --- */
  hfn = hfn_create(nnpar.width, nnpar.height);
  if (!hfn) error(E_NOMEM);     /* create a Hopfield network */

  /* --- show window --- */
  XtRealizeWidget(w_top);       /* show window on display */
  display = XtDisplay(w_top);   /* get display of top window */
  window  = XtWindow(w_net);    /* and drawable for window */
  if (db_alert(w_top, dcb_nop, NULL) != 0)
    error(E_WIDGET);            /* create alert dialog box */

  /* --- override window manager delete --- */
  XtOverrideTranslations(w_top, /* override window manager delete */
    XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
  wm_delwin = XInternAtom(XtDisplay(w_top), "WM_DELETE_WINDOW", False);
  XSetWMProtocols(XtDisplay(w_top), XtWindow(w_top), &wm_delwin, 1);

  /* --- allocate colors --- */
  screen = DefaultScreen(display);
  colmap = DefaultColormap(display, screen);
  black  = BlackPixel(display, screen);
  white  = WhitePixel(display, screen);
  if (XAllocNamedColor(display, colmap, "grey48", xc+1, xc) == 0)
    error(E_COLOR);             /* allocate colors and */
  grey   = xc[1].pixel;         /* note color pixel values */
  rseed((unsigned)time(NULL));  /* init. random number generator */

  /* --- evaluate command line arguments --- */
  if (argc > 1) dcb_load(NULL, (XtPointer)1, (XtPointer)argv[1]);

  /* --- start event loop --- */
  XtAppMainLoop(appctx);        /* enter main event loop */
  return 0;                     /* return 'ok' */
}  /* main() */
